package main

import "fmt"
import "math"
import "math/bits"
import "runtime" // only to get name of costfunction
import "reflect" // only to get name of costfunction
import "strings"
import "math/big" // used in SubsetsSizeK for error checking




// type definition: costfunction_t is another name for "func(*Tree, *Tree, QueryGraph)float64"
type costfunction_t func(*Tree, *Tree, QueryGraph)float64




/*
Representation of a Tree
*/
type Tree struct {
  Card     float64
  Rels     uint
  Left     *Tree
  Right    *Tree
  Cost     float64
  Costfunc costfunction_t
}

func (T *Tree) ToString () string {
  var isBaseRel bool = (T.Rels & (T.Rels-1) == 0) // is base relation if is power of two (exactly one bit set)
  if isBaseRel {
    return fmt.Sprint(bits.TrailingZeros(T.Rels))
  } else {
    res := "("
    if T.Left != nil {
      res += T.Left.ToString()
    }

    res += " "
    if T.Costfunc == nil {
      res += "x"
    } else {
      res += GetJoinImplNameFromCostFunc(T.Costfunc)
    }

    res += " "
    if T.Right != nil {
      res += T.Right.ToString()
    }
    res += ")"
    return res
  }
}


func mergeTrees (T1 *Tree, T2 *Tree, QG QueryGraph, costfunc costfunction_t) *Tree {
  var parent Tree
  parent.Card = T1.Card * T2.Card * QG.GetUnionSelTwoSets(T1.Rels, T2.Rels)
  parent.Rels = T1.Rels | T2.Rels
  parent.Left = T1
  parent.Right = T2
  parent.Cost = costfunc(T1, T2, QG)
  parent.Costfunc = costfunc
  return &parent
}



/*
Representation of a query graph:
- a slice of relations named R
- a map from uint to float for the join selectivities. Named S:
  - Let idx be a uint. Set bit i and j in idx to 1. Assign a selectivity to the S[idx]
*/
type QueryGraph struct {
  R []uint
  S map[uint]float64
}

func (QG *QueryGraph) SetSel (relA uint, relB uint, sel float64) {
  QG.S[1 << relA | 1 << relB] = sel
}

func (QG *QueryGraph) GetSel (relA uint, relB uint) float64 {
  val, ok := QG.S[1 << relA | 1 << relB]
  if (ok) {
    return val
  } else{
    return 1.0
  }
}


func (QG *QueryGraph) GetAllSels (Srels uint, rel uint) float64 {
  var sel float64 = 1.0
  for Srels != 0 {
    idx := uint(bits.TrailingZeros(Srels)) // number of trailing zeros == position of least significant bit set
    sel *= QG.GetSel(idx, rel)
    Srels &= ^(1 << idx) // "unary" ^ operator flips all bits
  }

  return sel
}

func (QG *QueryGraph) GetUnionSelTwoSets (Srels uint, Rrels uint) float64 {
  var sel float64 = 1.0
  for key := range QG.S {
    var matchIdxS uint = key & Srels
    var matchIdxR uint = key & Rrels
    if matchIdxS != 0 && matchIdxR != 0 {
      idx := matchIdxS | matchIdxR
      s, ok := QG.S[idx]
      if ok {
        sel *= s
      }
    }
  }
  return sel
}

func (QG *QueryGraph) Connected (R1 uint, R2 uint) bool {
  for key := range QG.S {
    if ((key & R1) != 0) && ((key & R2) != 0) {
      return true
    }
  }
  return false
}


func (QG *QueryGraph) SubsetsSizeK (k uint) []uint {
  if k < 1 {
    fmt.Println("k cannot be smaller than 1")
    return []uint{}
  }

  // believe it or not
  var n uint        = uint(len(QG.R))
  var firstsub uint = (1 << k) - 1
  var lastsub  uint = firstsub << (n - k)

  var res []uint
  res = append(res, firstsub)
  var sub uint = firstsub
  for sub != lastsub {
    low := uint(bits.TrailingZeros(sub)) // index of lowest bit set
    sub += (1 << low)
    high := uint(bits.TrailingZeros(sub))
    sub |= (uint(1) << (high - low - 1)) - 1
    res = append(res, sub)
  }

  // note: len(res) == BinomialCoefficient(n, k)
  // check if correct number of values is found
  var B big.Int // Package big implements arbitrary precision arithmetic (big numbers)
  (&B).Binomial(int64(n), int64(k))
  if uint64(len(res)) != B.Uint64() {
    panic("SubsetsSizeK(" + fmt.Sprint(k) + ") did not compute the correct number of subsets.")
  }

  return res

  /*
  * alternative way of doing it:
  * https://graphics.stanford.edu/~seander/bithacks.html
  * scrall down all to look at #NextBitPermutation
  */

  /*
  * If you think this is easy, send a detailed explanation of why the code works to the TA.
  * Your explanation will be included in the next exercise sheet.
  */
}



func (QG *QueryGraph) ToString () string {
  res := "Rels: " + fmt.Sprint(QG.R) + "\n"
  fmt.Println("Sels:")
  for rels, sel := range QG.S {
    res += fmt.Sprintf("%0" + fmt.Sprint(len(QG.R)) + "b", rels) + " : " + fmt.Sprintf("%.5f", sel) + "\n"
  }
  return res
}




// --- free functions ---
func Subsets (S uint) []uint {
  // return subsets of S
  // S itself and the empty set are not included
  // see script near DP-Bushy for details
  var res []uint
  S1 := S & (-S)
  for ok := true; ok; ok = (S1 != S) { // golang style for: do {...} while (S1 != S)
    res = append(res, S1)
    S1 = S & (S1 - S)
  }
  return res
}


func SetMinus (S1 uint, S2 uint, length uint) uint {
  mask := uint((1 << length) - 1)
  temp := S1 & S2
  temp = ^temp
  temp &= mask
  res := S1 & temp
  return res
}


// is S1 subset of S2?
func isSubset (S1 uint, S2 uint) bool {
  return ((S1 & S2) == S1)
}


func IdxsOfSetBits (S uint) []uint {
  // get the Indices of set bits in a uint (modelling a set)
  var res []uint
  for S != 0 {
    idx := uint(bits.TrailingZeros(S))
    S &= ^(1 << idx)
    res = append(res, idx)
  }
  return res
}


func IdxsOfUnsetBits (S uint, length uint) []uint {
  // get the Indices of unset bits in a uint (modelling a set)
  mask := uint((1 << length) - 1)
  Scomplement := (^S) & mask
  return IdxsOfSetBits(Scomplement)
}


func ValuesOfSetBits (S uint) []uint {
  Res := IdxsOfSetBits(S)
  for idx, idxval := range Res {
    Res[idx] = 1 << idxval
  }
  return Res
}


func ValuesOfUnsetBits (S uint, length uint) []uint {
  Res := IdxsOfUnsetBits(S, length)
  for idx, idxval := range Res {
    Res[idx] = 1 << idxval
  }
  return Res
}




// --- JoinTreeCreator ---
type JoinTreeCreator struct {
  RightDeepOnly bool
  LeftDeepOnly  bool
  Costfunctions []costfunction_t
}

func (JTC JoinTreeCreator) CreateJoinTree(T1 *Tree, T2 *Tree, QG QueryGraph) *Tree {
  // This implementation avoids cross products if possible
  if (QG.Connected(T1.Rels, T2.Rels)) { // are the relations of T1 and T2 connected in QG?
    mincost := math.Inf(1) // positive infinity
    var costfuncidx costfunction_t
    var left bool // join(T1, T2)  or  join(T2, T1)
    for _, impl := range JTC.Costfunctions {
      if !JTC.RightDeepOnly {
        thiscost := impl(T1, T2, QG)
        if mincost > thiscost {
          mincost = thiscost
          costfuncidx = impl
          left = true
        }
      }
      if !JTC.LeftDeepOnly {
        thiscost := impl(T2, T1, QG)
        if mincost > thiscost {
          mincost = thiscost
          costfuncidx = impl
          left = false
        }
      }
    }
    card := T1.Card * T2.Card * QG.GetUnionSelTwoSets(T1.Rels, T2.Rels)
    rels := T1.Rels | T2.Rels
    if left {
      return &Tree{card, rels, T1, T2, mincost, costfuncidx}
    } else {
      return &Tree{card, rels, T2, T1, mincost, costfuncidx}
    }
  } else { // cross product
    card := T1.Card * T2.Card
    rels := T1.Rels | T2.Rels
    cost := T1.Card * T2.Card + T1.Cost + T2.Cost
    return &Tree{card, rels, T1, T2, cost, Ccross}
  }
}




// --- cost functions ---
func Cout (T1 *Tree, T2 *Tree, QG QueryGraph) float64 {
  joincost := T1.Card * T2.Card * QG.GetUnionSelTwoSets(T1.Rels, T2.Rels)
  childrencost := T1.Cost + T2.Cost
  return joincost + childrencost
}


func Cnlj (T1 *Tree, T2 *Tree, QG QueryGraph) float64 {
  joincost := T1.Card * T2.Card
  childrencost := T1.Cost + T2.Cost
  return joincost + childrencost
}


func Chj (T1 *Tree, T2 *Tree, QG QueryGraph) float64 {
  h := 1.2
  joincost := h * T1.Card
  childrencost := T1.Cost + T2.Cost
  return joincost + childrencost
}


func Csmj (T1 *Tree, T2 *Tree, QG QueryGraph) float64 {
  joincost := T1.Card * math.Log2(T1.Card)  +  T2.Card * math.Log2(T2.Card)
  childrencost := T1.Cost + T2.Cost
  return joincost + childrencost
}


func Ccross (T1 *Tree, T2 *Tree, QG QueryGraph) float64 {
  joincost := T1.Card * T2.Card
  childrencost := T1.Cost + T2.Cost
  return joincost + childrencost
}


func applyCostfuncRec (T1 *Tree, T2 *Tree, QG QueryGraph, costfunc costfunction_t) float64 {
  // ToDo
  return -1.0
}



// --- helper method for convenience. Absolutely no need to understand any of this ---
func GetJoinImplNameFromCostFunc(impl costfunction_t) string {
  // ValueOf returns a new Value initialized to the concrete value stored in the interface i. ValueOf(nil) returns the zero Value. 
  var val reflect.Value = reflect.ValueOf(impl)
  // For Func arguments, Pointer returns a uintptr that is the underlying code pointer.
  var ptr uintptr = val.Pointer()
  // Package runtime contains operations that interact with Go's runtime system
  // FuncForPC returns a pointer to a Func object describing the function that contains the given program counter address
  var f *runtime.Func = runtime.FuncForPC(ptr)
  // Name() returns the name of the underlying function
  // result is: "packagename.functionname"
  var funcname string = f.Name()
  // split funcname, we are not interested in package name
  // convention: costfunction names are "C" followed by join implementation name
  split := strings.Split(funcname, ".C")
  joinimplname := split[1] // split[0] contains package name
  return joinimplname
}
